#include "../shared.h"
#include "randomnumbers.h"
#include "program.h"

__kernel void clear( write_only image2d_t outimg )
{
	int column = get_global_id( 0 );
	int line = get_global_id( 1 );
	if ((column >= SCRWIDTH) || (line >= SCRHEIGHT)) return;
	write_imagef( outimg, (int2)(column, line), 0 );
}

__kernel void clearGrid( global unsigned int* grid )
{
	int idx = get_global_id( 0 );
	int baseIdx = idx * (BALLSPERCELL + 1);
	grid[baseIdx] = 0;
}

__kernel void fillGrid( global float* balls, global unsigned int* grid )
{
	int idx = get_global_id( 0 );
	int gx = balls[idx * 4 + 0] * GRIDX;
	int gy = balls[idx * 4 + 1] * GRIDY;
	if ((gx < 0) || (gy < 0) || (gx >= GRIDX) || (gy >= GRIDY)) return;
	unsigned int baseIdx = (gx + gy * GRIDX) * (BALLSPERCELL + 1);
	if (grid[baseIdx] < BALLSPERCELL)
	{
		unsigned int count = atomic_inc( grid + baseIdx );
		if (count < BALLSPERCELL) grid[baseIdx + count + 1] = idx;
	}
	else balls[idx * 4 + 1] = balls[idx * 4 + 3] = 0.1;
}

__kernel void simulate1( global float* balls, float gx, float gy )
{
	int idx = get_global_id( 0 );
	float2 prevPos = { balls[idx * 4 + 0], balls[idx * 4 + 1] };
	float2 delta = { balls[idx * 4 + 0] - balls[idx * 4 + 2] + gx,
					 balls[idx * 4 + 1] - balls[idx * 4 + 3] + gy };
	float speed = length( delta );
	if (speed > 0.01f) delta = 0.01f * normalize( delta );
	balls[idx * 4 + 0] += delta.x * 0.95;
	balls[idx * 4 + 1] += delta.y * 0.95;
	balls[idx * 4 + 2] = prevPos.x;
	balls[idx * 4 + 3] = prevPos.y;
}

__kernel void simulateX( global float* balls, global read_only float* balls2, global read_only unsigned int* grid, float mx, float my )
{
	int ballIdx1 = get_global_id( 0 );
	float2 ball1Pos = { balls[ballIdx1 * 4 + 0], balls[ballIdx1 * 4 + 1] };
	int gx = ball1Pos.x * GRIDX, gx1 = max( 0.f, (ball1Pos.x - BALLRADIUS) * GRIDX ), gx2 = min( GRIDX - 1.f, (ball1Pos.x + BALLRADIUS) * GRIDX );
	int gy = ball1Pos.y * GRIDY, gy1 = max( 0.f, (ball1Pos.y - BALLRADIUS) * GRIDY ), gy2 = min( GRIDY - 1.f, (ball1Pos.y + BALLRADIUS) * GRIDY );
	if ((gx < 0) || (gy < 0) || (gx >= GRIDX) || (gy >= GRIDY)) return;
	float2 force = { 0, 0 };
	for (int gy = gy1; gy <= gy2; gy++) for (int gx = gx1; gx <= gx2; gx++)
	{
		int baseIdx = (gx + gy * GRIDX) * (BALLSPERCELL + 1);
		int count2 = min( (unsigned int)BALLSPERCELL, grid[baseIdx] );
		for (int j = 0; j < count2; j++)
		{
			int ballIdx2 = grid[baseIdx + j + 1];
			if (ballIdx2 != ballIdx1)
			{
				float2 ball2Pos = { balls2[ballIdx2 * 4 + 0], balls2[ballIdx2 * 4 + 1] };
				float2 delta = ball2Pos - ball1Pos;
				float dist = delta.x * delta.x + delta.y * delta.y;
				if (dist < (BALLRADIUS * BALLRADIUS))
				{
					dist = native_sqrt( dist );
					float2 push = (BALLRADIUS - dist) * PUSHFORCE * (delta / dist);
					force -= push;
				}
			}
		}
	}
	// evade mouse
	float2 delta = { ball1Pos.x - mx, ball1Pos.y - my };
	float dist = delta.x * delta.x + delta.y * delta.y;
	if (dist < (0.05f * 0.05f))
	{
		dist = native_sqrt( dist );
		float2 push = (0.05f - dist) * PUSHFORCE * (delta / dist);
		force += push;
	}
	// limit impulse
	float speed = length( force );
	if (speed > MAXFORCE) force = MAXFORCE * normalize( force );
	ball1Pos += force;
	// respect boundaries
	float jitter = (float)ballIdx1 * 0.000001f;
	ball1Pos.x = min( 0.98f + jitter, max( 0.02f - jitter, ball1Pos.x ) );
	ball1Pos.y = min( 0.98f + jitter, max( 0.02f - jitter, ball1Pos.y ) );
	// update ball position
	balls[ballIdx1 * 4 + 0] = ball1Pos.x;
	balls[ballIdx1 * 4 + 1] = ball1Pos.y;
}

__kernel void render( write_only image2d_t outimg, global read_only float* balls, global read_only unsigned int* grid )
{
	int column = get_global_id( 0 );
	int line = get_global_id( 1 );
	if ((column >= SCRWIDTH) || (line >= SCRHEIGHT)) return;
	float2 uv = { (float)column / SCRWIDTH, (float)line / SCRHEIGHT };
	// draw balls using grid
	int gx = uv.x * GRIDX;
	int gy = uv.y * GRIDY;
	int gx1 = max( 0, gx - 1 ), gx2 = min( GRIDX - 1, gx + 1 );
	int gy1 = max( 0, gy - 1 ), gy2 = min( GRIDY - 1, gy + 1 );
	for (int gy = gy1; gy <= gy2; gy++) for (int gx = gx1; gx <= gx2; gx++)
	{
		unsigned int baseIdx = (gx + gy * GRIDX) * (BALLSPERCELL + 1);
		unsigned int count = min( (unsigned int)BALLSPERCELL, grid[baseIdx] );
		for (int i = 0; i < count; i++)
		{
			unsigned int idx = grid[baseIdx + i + 1];
			float2 delta = { balls[idx * 4] - uv.x, balls[idx * 4 + 1] - uv.y };
			float dist = delta.x * delta.x + delta.y * delta.y;
			if (dist < (BALLRENDERSIZE * BALLRENDERSIZE))
			{
				write_imagef( outimg, (int2)(column, (SCRHEIGHT - 1) - line), (float4)(1, 0, 0, 1) );
				break;
			}
		}
	}
}